# html and text-to-html filters

package Ewok::Filter::HTML;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = ();
@EXPORT_OK = qw(
);
%EXPORT_TAGS = (
    all => [ @EXPORT_OK ],
);
use Ewok::Config;
use Ewok::Filter qw(:register default_pipeline);
use strict;

# Register filters and pipelines with Ewok::Filter
BEGIN {
    register_filter(
        html_heading_anchors => \&html_heading_anchors,
        text_indented_sections => \&text_indented_sections,
        text_escape_html => \&text_escape_html,
        text_tables => \&text_tables,
        text_headings => \&text_headings,
        text_hr => \&text_hr,
        text_paragraph => \&text_paragraph,
        text_break_short_lines => \&text_break_short_lines,
        text_cleanup_pre_sections => \&text_cleanup_pre_sections,
        text_paragraph_classes_styles => \&text_paragraph_classes_styles,
        text_bold_italics => \&text_bold_italics,
        text_links => \&text_links,
        text_embperl_cleanup => \&text_embperl_cleanup,
        html_named_anchors => \&html_named_anchors,
        html_headings => \&html_headings,
        html_contents => \&html_contents,
    );
    register_pipeline(
        StandardHTML => [qw(
            html_heading_anchors
            html_named_anchors html_headings html_contents
        )],
        StandardTextHTML => [qw(
            html_heading_anchors
            text_escape_html text_indented_sections text_headings text_hr 
            text_paragraph text_break_short_lines text_cleanup_pre_sections 
            text_paragraph_classes_styles text_bold_italics text_links 
            text_embperl_cleanup
            html_named_anchors html_headings html_contents
        )],
        MinimalTextHTML => [qw(
            html_heading_anchors
            text_headings text_hr
            text_paragraph text_break_short_lines text_cleanup_pre_sections 
            text_embperl_cleanup
            html_named_anchors html_headings
        )],
    );
    default_pipeline('StandardTextHTML');
}

# Add named anchors to html headings
sub html_heading_anchors
{
    local $_ = shift;
    my $arg = shift;
    my $markup = $arg->{markup_heading_anchors};
    return $_ if defined $markup && $markup == 0;

    # Escape quotes in headings
    while (s|(<h([1-6])>.*?)"(.*</h\2>)|$1&quot;$3|img) { 1 }; #"
    # <hX>Foo</hX> => <hX>"Foo"=#foo</hX>
    s|(<h([1-6])[^>]*>)\s*(.+?)\s*(</h\2>)(\r?\n)?|$1"$3"=#\L$3\E$4\n|img;

    return $_;
}

# Close any outstanding pre sections/lists up to given $indent
sub sections_close
{
    my ($output, $newindent, $indent, $type, $marker) = @_;
    my $type_prev = '';

    while ($newindent < $indent->[0]) {
        my $type = shift @$type;
        if ($type eq 'pre') {
            $$output .= "</$type>\n\n";
        } else {
            $$output .= ' ' x $indent->[0];
            $$output .= '</pl>' if $type_prev ne 'pre';
            $$output .= "</li>\n";
            $$output .= ' ' x $indent->[0] . "</$type>\n";
        }
        shift @$indent;
        shift @$marker;
        $type_prev = $type;
    }
}

# Various cleanups from text_indented_sections formatting
sub sections_cleanup
{
    my $output = shift;

    # Remove empty paragraphs
    $$output =~ s!<pl>\s*</pl>!!ig;
    # Remove extraneous </p> tags beginning lines
    $$output =~ s!^(\s*)</pl>\n?!$1!imsg;
    # Remove trailing empty lines in <pre> sections
    $$output =~ s!\n\s*(\n</pre>)!$1!isg;
    # Remove trailing whitespace in paragraphs
    $$output =~ s!\s+(</pl>)!$1!ig;
    # Remove paragraphs from simple <li> .. </li> lines
    $$output =~ s!(<li>)<pl>(.*?)</pl>(</li>)!$1$2$3!ig;
    my $out = $$output;
    while ($$output =~ m!<li><pl>(.*?)</pl></li>!isg) {
        my $content = $1;
        $out =~ s!(<li>)<pl>(\Q$content\E)</pl>(</li>)!$1$2$3!is
            if $content !~ m!</?pl>!i;
    }
    $$output = $out unless $out eq $$output;
    # Finally, map our <pl> paragraphs back to normal <p> ones
    $$output =~ s!(</?)pl(>)!$1p$2!ig;
}

# Markup lists and <pre> sections (procedural, not regex-based)
sub text_indented_sections
{
    my $text = shift;
    my $arg = shift;

    my $orderedMarker = qr/N[AaIi]?\./;
    my $unorderedMarker = qr/[-\*o\+]/;
    my $listMarker = qr/$orderedMarker|$unorderedMarker/;
                
    my $output = '';
    my @indent = (-1);
    my @type = ('');
    my @marker = ('');
    my $empty_line = 0;
    my $extraindent = 0;

    # Iterate over lines
    for (split /\r?\n/, $text) {
        my $marker = '';

        # Skip empty lines
        if (m/^\s*$/) {
            # In non-indented or <pre> sections, just pass empty lines through
            if (! $type[0] || $type[0] eq 'pre') {
                $output .= "$_\n";
            }
            # Elsewhere, just note empty lines to close off current paragraphs
            else {
                $empty_line = 1;
            }
            next;
        }

        # Get newindent
        my $newindent = (m/^(\s+)/) ? length($1) : 0;

        # Decreased indentation - close hanging tags
        if ($newindent < $indent[0]) {
            sections_close(\$output, $newindent, \@indent, \@type, \@marker);
        }

        # If no indent or within a <pre> section, pass straight through
        if ($newindent == 0 || $type[0] eq 'pre') {
            $output .= "$_\n";
            next;
        }

        # Indented line - check for initial list marker
        if (m/^(\s+)($listMarker)(\s+)(.+?)\s*$/) {
            my $marker = $2;
            my $type = substr($marker,0,1) eq 'N' ? 'ol' : 'ul';
            $extraindent = $newindent + length($2) + length($3);
            my $item = $4;
            $empty_line = 0;

            # If increased indentation, must be nested list (if started)
            if ($newindent > $indent[0]) {
                # Terminate current
                my $nested = $type[0] && $type[0] ne 'pre';
                if ($nested) {
                  $output .= ' ' x $indent[0] . "</pl></li>\n";
                  $output .= ' ' x $indent[0] . "<li>\n";
                }
                unshift @indent, $newindent;
                unshift @type, $type;
                unshift @marker, $marker;
                $output .= ' ' x $indent[0];
                $output .= "<$type>\n";
                $output .= ' ' x $indent[0] . "<li><pl>$item";
            }

            # If type or marker changed, begin a new list at this level
            elsif ($type ne $type[0] || $marker ne $marker[0]) {
                # Type or marker changed - begin a new list, same level
                $output .= ' ' x $indent[0] . "</pl></li>\n";
                $output .= ' ' x $indent[0] . "</$type[0]>\n";
                $type[0] = $type;
                $marker[0] = $marker;
                $output .= ' ' x $indent[0] . "<$type>\n";
                $output .= ' ' x $indent[0] . "<li><pl>$item";
            }

            # If same indentation, must be same list
            else {
                # Terminate current
                $output .= ' ' x $indent[0] . "</pl></li>\n";
                $output .= ' ' x $indent[0] . "<li><pl>$item";
            }
        }
        
        # Not pre, not list item - check if paragraph within current list
        elsif ($type[0] && m/^\s{$extraindent}(\S.*)/) {
            if ($empty_line) {
                $output .= "</pl>\n" . ' ' x $indent[0] . "<pl>";
                $empty_line = 0;
            }
            else {
                $output .= "\n" . ' ' x $indent[0];
            }
            $output .= $1;
        }

        # Not pre, not list, not current paragraph - must be a new pre
        else {
            $output .= "</pl>\n" if $type[0];
            unshift @indent, $newindent;
            unshift @type, 'pre';
            unshift @marker, '';
            $output .= "<pre>\n";
            $output .= "$_\n";
        }
    }

    # Close off hanging tags
    sections_close(\$output, -1, \@indent, \@type, \@marker);

    # Various cleanups 
    sections_cleanup(\$output);

    return $output;
}

# Escape special HTML characters
sub text_escape_html 
{
    local $_ = shift;
    my $arg = shift;
    return $_ if defined $arg->{escape_html} && 
        $arg->{escape_html} == 0;

    # Escape html tags
    s|<|&lt;|g;
    s|>|&gt;|g;
 
    return $_;
}

# Markup text tables of form |a|b|c|d|
sub text_tables
{
    local $_ = shift;
    my $arg = shift;
    my $markup_classes_styles = $arg->{markup_classes_styles};
    $markup_classes_styles = 1 if ! defined $arg->{markup_classes_styles};

    return $_ if defined $arg->{markup_tables} && 
        $arg->{markup_tables} == 0;

    my ($class, $style);
    if ($markup_classes_styles) {
        $class = '(?<!\\\\)\[\s*([^\]]*?)\s*\]\s*';
        $style = '(?<!\\\\){\s*([^\}]*?)\s*\}\s*';
    }
    my $text = '';
    for my $line (split /\r?\n/, $_) {
        # Begin-table tag (with support for classes and styles)
        if ($line =~ m/^[ \t]*($class)?($style)?($class)?($style)?(?<!\\)\|(.*)$/) {
            $line = '<table';
            if ($6 || $8) {
                $line .= qq( class="$2") if $2;
                $line .= qq( style="$4") if $4;
            }
            $line .= '><tr';
            if ($6 || $8) {
                $line .= qq( class="$6") if $6;
                $line .= qq( style="$8") if $8;
            }
            # If single set assume <tr> over <table>
            else {
                $line .= qq( class="$2") if $2;
                $line .= qq( style="$4") if $4;
            }
            $line .= "><td>$9";
        }
        $text .= "$line\n";
    }
    $_ = $text;
    # End-table tag 
    s:(?<!\\)\|\s*?$:</td></tr></table>:mg;
    # Between table cells
    s:(?<!\\)\|:</td><td>:g;
    # Markup <td> classes and styles
    s:<td([^>]*)>\s*$class:<td$1 class="$2">:g if $class;
    s:<td([^>]*)>\s*$style:<td$1 style="$2">:g if $style;
    # Remove extraneous <table> and </table> tags
    s|</table>(.*?\n.*?)<table[^>]*>|$1|g;
    # &nbsp; empty cells
    s|(<td[^>]*>)(</td>)|$1&nbsp;$2|g;
    # If <table> not deleted and single set, assume <table> over <tr>
    s|<table>\s*<tr(\s+class=".*?")?(\s+style=".*?")?\s*>|<table$1$2><tr>|g;
    # Border pseudoclass - translate a table [border] class to border="1"
    s|<table class="border"|<table border="1"|ig;
    # th pseudoclass - translate a tr [th] class to use <th> tags
    while (s|(<tr class="th".*?)(</?)td|$1$2th|im) { 1 };
    s|<tr class="th"|<tr|ig;

    return $_;
}

# Text headings. Requires that 'html_named_anchors' is called subsequently to
#   expand "Heading"=#heading constructs.
sub text_headings
{
    local $_ = shift;
    my $arg = shift;
    $arg->{markup_heading_anchors} = 1 
        if ! defined $arg->{markup_heading_anchors};
    
    # Markup headings - underlined headings, h1 to h3
    my $level = { '=' => 1, '-' => 2, '~' => 3 };
    s|^[ \t]*(\S.*)\n[ \t]*([-=~])[-=~]{2,}\s*?$|h$level->{$2}: $1|mg;
    if ($arg->{markup_heading_anchors}) {
        # h1: Foo =#foofoo
        s|^[ \t]*(h[1-6]):[ \t]*(.+?)\s*(=#\S+)\s*$|<$1>"$2"$3</$1>\n|img;
        # h1: Foo
        s|^[ \t]*(h[1-6]):[ \t]*([^:].*?)\s*$|<$1>"$2"=#\L$2\E</$1>\n|img;   #"
        # h1:: Foo             #"
        s|^[ \t]*(h[1-6])::[ \t]*(.+?)\s*$|<$1>$2</$1>\n|img;
    }
    else {
        # h1: Foo             #"
        s|^[ \t]*(h[1-6])::?[ \t]*(.+?)\s*(=#\S+\s*)?$|<$1>$2</$1>\n|img;
    }

    return $_;
}

# Markup HRs
sub text_hr
{
    local $_ = shift;

    # HRs
    s|^[-_]{4,}\s*?$|<hr>|mg;
 
    return $_;
}
    
# Text to HTML paragraphs
sub text_paragraph
{
    local $_ = shift;
    return $_ if m/^\s*$/s;

    # Add an opening <p> at the very beginning
    s|^\s*|<pp>|;
    # Add a closing <p> at the very end
    s|\s*$|</pp>\n|;
    # Add paragraph tags at all empty lines
    s|\r?\n\s*(\r?\n)|</pp>$1<pp>|g;
    # Add an open <p> after lists and pre sections
    s!(</(pre|ul|ol)>\r?\n)(\S)!$1<pp>$3!g;

    # Remove any final hanging <p> tag
    s|<pp>\s*$||;
    # Remove any tags around headings
#   s|^<pp>(\s*"?\s*<h\d)|$1|img;   #"
#   s|(</h\d>.*)</pp>$|$1|img;
    # Remove any tags around lists + tables
#   s|^<pp>(\s*<table)|$1|img;
#   s|(</table>\s*)</pp>$|$1|img;
    # Remove all paragraph tags enclosing other tags (except <b>,<u>,<i>)
    s|^<pp>(\s*<(?![biu]\>))|$1|img;
    s|(?<!</[biu])>\s*</pp>$|>|img;

    # Finally, munge our <pp> tags to proper <p> ones
    s!(</?)pp(>)!$1p$2!ig;

    return $_;
}

# Break short lines
sub text_break_short_lines
{
    local $_ = shift;
    my $arg = shift;
return $_;
    my $length = exists $arg->{length} ? $arg->{length} : 40;
    return $_ if $length == 0;

    # Add <br /> tags on any short (<$length chars) lines 
    #   (not ending in > or embperl [-+#\$!]\]
    $length--;
    s:^(\s*\S.{0,$length}(?<!.>|[-+#\$!]\]))\s*?$:$1<br />:mg;

    return $_;
}

# Remove <br /> and <p> tags within <pre></pre> sections
sub text_cleanup_pre_sections
{
    local $_ = shift;
    return $_ unless m/<pre>/i;

    my $new = '';
    for my $chunk (split m!</pre>!ism) {
        my ($rest, $pre) = split m!<pre>!ism, $chunk, 2;
        # Remove <p> and <br /> tags
        if ($pre) {
            $pre =~ s!</p\b[^>]*>!\n!ig;
            $pre =~ s!<p\b[^>]*>!!ig;
            $pre =~ s!<br\b[^>]*/?>!!ig;
        }
        # Add this chunk to $new
        $new .= $rest if $rest;
        $new .= "<pre>$pre</pre>" if $pre;
    }

    return $new;
}

# [class] and {style} markup for paragraphs
sub text_paragraph_classes_styles
{
    local $_ = shift;
    my $arg = shift;
    return $_ if defined $arg->{markup_classes_styles} &&
        $arg->{markup_classes_styles} == 0;

    s!^<p>\s*\[\s*(\S.*?)\s*\]\s*!<p class="$1">!img;
    s!^<p([^>]*)>\s*{\s*(\S.*?)\s*\}\s*!<p$1 style="$2">!img;

    return $_;
}

# Markup bold and italics elements
sub text_bold_italics
{
    local $_ = shift;
    my $arg = shift;
    $arg->{markup_bold} = 1 if ! defined $arg->{markup_bold};
    $arg->{markup_italics} = 1 if ! defined $arg->{markup_italics};

    # Add formatting tags (strong, em, etc)
    s|(?<!\\)\*(\S(.*?\S)??)(?<!\\)\*|<strong>$1</strong>|mg
        if $arg->{markup_bold};
    s|(?<!\\)__(\S(.*?\S)??)(?<!\\)__|<strong><em>$1</em></strong>|mg
        if $arg->{markup_bold} && $arg->{markup_italics};
    s|(?<!\\)_(\S(.*?\S)??)(?<!\\)_|<em>$1</em>|mg
        if $arg->{markup_italics};

    # Unescape any escaped formatting tags
    s|\\([*_\|\[\{])|$1|mg;

    return $_;
}

# Markup links
sub text_links
{
    local $_ = shift;
    my $arg = shift;
    return $_ if defined $arg->{markup_links} && $arg->{markup_links} == 0;

    # Markup mailto links
    # Labelled mailto
    s!"([^"]+)":\s*(mailto:\w[-.+\w]+\@([-\w]+\.)+\w+(\?[^\s<]*[^\s<\.])?) #"
        !<a href="$2">$1</a>!imgx;
    # Labelled email
    s!"([^"]+)":\s*(\w[-.+\w]+\@([-\w]+\.)+\w+)  #"
        !<a href="mailto:$2">$1</a>!imgx;
    # Bare mailto
    s|(?<!\")(mailto:(\w[-.+\w]+\@([-\w]+\.)+\w+)(\?[^\s<]*[^\s<\.])?)  #"
        |<a href="$1">$2</a>|imgx;
    # Bare email in <>
    s! ([^:\w]) (<|&lt;) (\w[\-\.\+\w]+\@([-\w]+\.)+\w+) (>|&gt;)  #"
        !$1<a href="mailto:$3">&lt;$3&gt;</a> !igx;
    # Bare email
    s! ([^:;\w]) (\w[-.+\w]+\@([-\w]+\.) +\w+)  #"
        !$1<a href="mailto:$2">$2</a>!imgx;

    # Markup http links
    # s!"([^"]+)":\s*(http://[^\s<]+[\w/][\w/\?=\+])    #"
    #         !<a href="$2">$1</a>!imgx;
    # s!"([^"]+)":\s*([^\s<]+[^\s<\.][\w/\?=\+])   #"
    #           !<a href="$2">$1</a>!imgx;
    # s|(?<!\")(http://[^\s<]+[^\s<\.][\w/\?=\+])     #"
    #           |<a href="$1">$1</a>|imgx;
    s!"([^"]+)":\s*([^\s<]+[\w/\?=\+])!<a href="$2">$1</a>!imgx;
    s|(?<!\")(http://[^\s<]+[\w/\?=\+])|<a href="$1">$1</a>|imgx;    #"
  
    return $_;
}

# Cleanup any wrongly formatted embperl code sections
sub text_embperl_cleanup
{
    local $_ = shift;
    my $arg = shift;
    return $_ unless $arg->{allow_syntax} && $arg->{allow_syntax} eq 'Text';

    my $new = '';
    my $code_open = qr((\[?\[[-+#!\$]));
    my $code_type;
    for (split $code_open) {
        if ($_ =~ $code_open) {
            ($code_type) = m/\[([-+#!\$])/;
        }
        else {
            if ($code_type) {
                my $b;
                ($_, $b) = split /\Q$code_type\E\]/, $_, 2;
                # <pre> tags on own line within code
                1 while s:(.*?\n)</?pre>\r?\n(.*):$1$2:isg;
                # <br /> at $ within code
                1 while s:(.*?)<br[ \t]*/?>(\r?\n.*):$1$2:isg;
                # <p> at ^ within code
                1 while s:(.*?\n)<p[ \t]*/?>(.*):$1$2:isg;
                # </p> at $ within code
                1 while s:(.*?)</p[ \t]*/?>(\r?\n.*):$1$2:isg;
                $_ = $_ . "$code_type]" . $b;
                # Reset $code_type
                $code_type = '';
            }
        }
        $new .= $_;
    }
    $_ = $new;
    # <p> at ^ before code
    s:^<p[\t ]*/?>(\[[-+#\$!].*)$:$1:img;
    # </p> at $ after code
    s:^(.*[-+#\$!]\])\s*</p>$:$1:img;

    return $_;
}

# Parse (and perhaps delete) top-level html headings
sub html_headings
{
    local $_ = shift;
    my $arg = shift;

    return $_ unless $arg->{delete_initial_heading} || 
            ref $arg->{h1} eq 'SCALAR' || 
            ref $arg->{h2} eq 'SCALAR' || 
            $arg->{auto_contents};

    my $delete_initial_heading = $arg->{delete_initial_heading};
    my $maxTags = 10;
    my ($lead, $stag, $etag);

    my $contents = '';
    my $Config = get_config();
    $contents = $Config->{EWOK_AUTO_CONTENTS_STRING}
        if $arg->{auto_contents} && ! m!((<|&lt;)contents\b([^>]*)>)!is;

    # Find first h1 tag. This is somewhat nasty - need to try to limit 
    #   the number of tags we will search through to try and find our heading
    my ($h1, $h1_full);
    if ($delete_initial_heading || ref $arg->{h1} eq 'SCALAR') {
        if (m|(<body.*?)(<h1\b[^>]*>)((?:<a name=[^>]+>)?(.*?)(?:</a>)?)(</h1.*?>)|is ||
                m|^(.*?)(<h1\b[^>]*>)((?:<a name=[^>]+>)?(.*?)(?:</a>)?)(</h1.*?>)|is) {
            $lead = $1;
            $stag = $2;
            $h1_full = $3;
            $h1 = $4;
            $etag = $5;
        }
        # Limit no of tags between opening <body> (or beginning) and <h1>
        my $tags = ($lead =~ s|<|<|og);
        undef $h1 if $tags > $maxTags;

        # Do the h1 delete
        if ($h1) {
            if ($delete_initial_heading) {
                s|(<p>)?\s*$stag\Q$h1_full\E$etag(\s*</p>)?(\s*\r?\n)?|$contents|is;
            }
            elsif ($contents) {
                s|((<p>)?\s*$stag\Q$h1_full\E$etag(\s*</p>)?(\s*\r?\n)?)|$1$contents|is;
            }
        }
    }

    # Find first h2 tag. 
    my ($h2, $h2_full);
    if (ref $arg->{h2} eq 'SCALAR') {
        if (m|(<body.*?)(<h2\b[^>]*>)((?:<a name=[^>]+>)?(.*?)(?:</a>)?)(</h2.*?>)|is ||
                m|^(.*?)(<h2\b[^>]*>)((?:<a name=[^>]+>)?(.*?)(?:</a>)?)(</h2.*?>)|is) {
            $lead = $1;
            $stag = $2;
            $h2_full = $3;
            $h2 = $4;
            $etag = $5;
        }
        # Limit no of tags between opening <body> (or beginning) and <h2>
        my $tags = ($lead =~ s|<|<|og);
        undef $h2 if $tags > $maxTags;

        # Delete the h2 if no h1 found
        if ($h2 && ! $h1) {
            if ($delete_initial_heading) {
                s|(<p>)?\s*$stag\Q$h2_full\E$etag(\s*</p>)?(\s*\r?\n)?|$contents|is;
            }
            elsif ($contents) {
                s|((<p>)?\s*$stag\Q$h2_full\E$etag(\s*</p>)?(\s*\r?\n)?)|$1$contents|is;
            }
        }
    }

    # Set h1 and h2 in $arg if scalar refs
    ${$arg->{h1}} = $h1 if $h1 && ref $arg->{h1} eq 'SCALAR';
    ${$arg->{h2}} = $h2 if $h2 && ref $arg->{h2} eq 'SCALAR';

    return $_;
}

#
sub html_named_anchors
{
    local $_ = shift;
#   my $arg = shift;

    # Markup named anchors
    # Remove spaces, punctuation, tags and entities from labels
    while (s|("[^"]*"=#\S*)[ \t]+|$1_|g) { 1; }               #"
    while (s|("[^"]*"=#\w*)</?[^h>/][^>]*>|$1_|ig) { 1; }        #"
    while (s|("[^"]*"=#\w*)&[^;]{2,6};|$1_|g) { 1; }          #"
    while (s|("[^"]*"=#\w*)[-!?"'/\\:;,.()@#\$%^&*[\]]+|$1_|g) { 1; }     #'
    while (s|("[^"]*"=#\w*?)__+|$1_|g) { 1; }                 #"
    # Remove leading and trailing underscores from labels
    s|("[^"]*"=#)_|$1|g;                                      #"
    s|("[^"]*"=#\w+)_(\W)|$1$2|g;                             #"
    # Markup labelled and unlabelled named anchors
    s|"([^"]*)"=#(\w[^ \t<]*)|<a name="$2">$1</a>|mg;       #"
    s|"([^"]*)"=#([^ \t<]+)|<a name="ref$2">$1</a>|mg;      #"
    s|^=#(\w[^ \t<]*)|<a name="$1"></a>|mg;
    s|^=#([^ \t<]+)|<a name="ref$1"></a>|mg;

    # Change headings from using named anchors to using ids
    s!<(h[1-6])><a name="([^"]+)">([^<]+)</a></\1>!<$1 id="$2">$3</$1>!img;

    return $_;
}

# <contents> taglib
sub html_contents
{
    local $_ = shift;
    my $arg = shift;
    my $markup_contents = $arg->{markup_contents};
    $markup_contents = 1 if ! defined $markup_contents;

    # Replace <contents> pseudotag
    my $contents_tag;
    if ($markup_contents && 
            (($contents_tag) = m!((<|&lt;)contents\b(.*?)(>|&gt;))!is)) {
        my $attr = $3;
        my $contents = '';
        my $type = 'ul';
        my $indent = 2;
        if ($attr =~ m/\btype="?(\w+)"?\b/i) {
            $type = $1;
        }
        if ($attr =~ m/\bindent="?(\d+)"?\b/i) {
            $indent = $1;
        }
        {
            # Ignore everything up to the contents tag
            local $_ = $_;
            s/^.*$contents_tag//s;
            my @headings = m!(<(h[1-6])\b.*?\</\2>)!ig;
            if (@headings) {
                my %indent = ();
                if ($indent) {
                    # Cache the heading levels found in %h
                    for my $h (@headings) {
                        $indent{$1} = 1 if ($h =~ m|<h([1-6])\b|);
                    }
                    my $i = 0;
                    for my $j (1 .. 6) {
                        $indent{$j} = $i++ if $indent{$j};
                    }
                }
                else {
                    $contents .= "<$type>\n" if $type ne 'simple';
                }
                $contents .= "<p>" if $type eq 'simple';
                my $lasthead = 0;
                for my $h (@headings) {
                    # Pick out named anchor label (if any) and heading
#                   if ($h =~ m|((?:<a name="([^"]+))?.*)?<h([1-6])\b[^>]*>(.+?)</h\3>|i) {
                    if ($h =~ m|<h([1-6])\b[^>]*>(<a name="([^"]+)">)?(.+?)(</a>)?</h\1>|i) {
                        my $head = $1;
                        if ($type eq 'simple') {
                            $contents .= '&nbsp;' x ($indent * $indent{$head});
                        }
                        else {
                            if ($head > $lasthead) {
                                $contents .= "<$type>";
                            }
                            elsif ($head < $lasthead) {
                                $contents .= "</$type>";
                            }
                            $contents .= '<li>';
                        }
                        if ($3) {
                            $contents .= qq(<a href="#$3">$4</a>);
                        } 
                        else {
                            $contents .= $4;
                        }
                        $contents .= ($type ne 'simple') ? "</li>\n" : "<br />\n";
                        $lasthead = $head;
                    }
                }
                if ($type eq 'simple') {
                    $contents .= "</p>";
                }
                else {
                    for (0 .. $indent{$lasthead}) {
                        $contents .= "</$type>\n";
                    }
                }
            }
        }
        # Escape regex special chars
        $contents_tag =~ s/([\\^\.\$\|\(\)\[\]])/\\$1/g;
        s|$contents_tag|$contents|i;
    }

    return $_;
}

1;

__END__

=head1 NAME

Ewok::Filter::HTML - various html and text-to-html filters


=head1 SYNOPSIS

    use Ewok::Filter::HTML qw(paragraph);

    $content = paragraph($content);



=head1 DESCRIPTION




=head1 AUTHOR

Gavin Carr <gavin@openfusion.com.au>


=head1 SEE ALSO

Ewok::Filter


=head1 COPYRIGHT

Copyright 2002-2003, Open Fusion Pty. Ltd., Australia. All Rights Reserved.

This program is free software. You may copy or redistribute it under the same 
terms as perl itself.

=cut


# arch-tag: f3334f14-c736-4623-a9be-a46a8e4d5be4
